@@ -59,54 +59,57 @@ extern mach_port_t _lookupd_port();
 static int gai_proc = -1;
 static int gni_proc = -1;
 
-static int32_t supported_family[] =
+static const int32_t supported_family[] =
 {
 	PF_UNSPEC,
 	PF_INET,
 	PF_INET6
 };
-static int32_t supported_family_count = 3;
+static const int32_t supported_family_count = (sizeof(supported_family) / sizeof(supported_family[0]));
 
-static int32_t supported_socket[] =
+static const int32_t supported_socket[] =
 {
 	SOCK_RAW,
 	SOCK_UNSPEC,
 	SOCK_DGRAM,
 	SOCK_STREAM
 };
-static int32_t supported_socket_count = 4;
+static const int32_t supported_socket_count = (sizeof(supported_socket) / sizeof(supported_socket[0]));
 
-static int32_t supported_protocol[] =
+static const int32_t supported_protocol[] =
 {
 	IPPROTO_UNSPEC,
+	IPPROTO_ICMP,
 	IPPROTO_ICMPV6,
 	IPPROTO_UDP,
 	IPPROTO_TCP
 };
-static int32_t supported_protocol_count = 4;
+static const int32_t supported_protocol_count = (sizeof(supported_protocol) / sizeof(supported_protocol[0]));
 
-static int32_t supported_socket_protocol_pair[] =
+static const int32_t supported_socket_protocol_pair[] =
 {
 	SOCK_RAW,    IPPROTO_UNSPEC,
 	SOCK_RAW,    IPPROTO_UDP,
 	SOCK_RAW,    IPPROTO_TCP,
+	SOCK_RAW,    IPPROTO_ICMP,
 	SOCK_RAW,    IPPROTO_ICMPV6,
 	SOCK_UNSPEC, IPPROTO_UNSPEC,
 	SOCK_UNSPEC, IPPROTO_UDP,
 	SOCK_UNSPEC, IPPROTO_TCP,
+	SOCK_UNSPEC, IPPROTO_ICMP,
 	SOCK_UNSPEC, IPPROTO_ICMPV6,
 	SOCK_DGRAM,  IPPROTO_UNSPEC,
 	SOCK_DGRAM,  IPPROTO_UDP,
 	SOCK_STREAM, IPPROTO_UNSPEC,
 	SOCK_STREAM, IPPROTO_TCP
 };
-static int32_t supported_socket_protocol_pair_count = 12;
+static const int32_t supported_socket_protocol_pair_count = (sizeof(supported_socket_protocol_pair) / (sizeof(supported_socket_protocol_pair[0]) * 2));
 
 static int
 gai_family_type_check(int32_t f)
 {
 	int32_t i;
-	
+
 	for (i = 0; i < supported_family_count; i++)
 	{
 		if (f == supported_family[i]) return 0;
@@ -119,7 +122,7 @@ static int
 gai_socket_type_check(int32_t s)
 {
 	int32_t i;
-	
+
 	for (i = 0; i < supported_socket_count; i++)
 	{
 		if (s == supported_socket[i]) return 0;
@@ -132,7 +135,7 @@ static int
 gai_protocol_type_check(int32_t p)
 {
 	int32_t i;
-	
+
 	for (i = 0; i < supported_protocol_count; i++)
 	{
 		if (p == supported_protocol[i]) return 0;
@@ -145,7 +148,7 @@ static int
 gai_socket_protocol_type_check(int32_t s, int32_t p)
 {
 	int32_t i, j, ss, sp;
-	
+
 	for (i = 0, j = 0; i < supported_socket_protocol_pair_count; i++, j+=2)
 	{
 		ss = supported_socket_protocol_pair[j];
@@ -465,7 +468,7 @@ gai_lookupd_process_dictionary(XDR *inxdr)
 static int
 gai_make_query(const char *nodename, const char *servname, const struct addrinfo *hints, char *buf, uint32_t *len)
 {
-	int32_t numerichost, family, proto, socktype, canonname, passive;
+	int32_t numerichost, family, proto, socktype, canonname, passive, parallel;
 	uint32_t na;
 	XDR outxdr;
 	char str[64], *cname;
@@ -476,6 +479,7 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 	socktype = SOCK_UNSPEC;
 	canonname = 0;
 	passive = 0;
+	parallel = 0;
 	cname = NULL;
 
 	if (hints != NULL)
@@ -483,7 +487,8 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 		family = hints->ai_family;
 		if (hints->ai_flags & AI_NUMERICHOST) numerichost = 1;
 		if (hints->ai_flags & AI_CANONNAME) canonname = 1;
-		if ((hints->ai_flags & AI_PASSIVE) == 1) passive = 1;
+		if (hints->ai_flags & AI_PASSIVE) passive = 1;
+		if (hints->ai_flags & AI_PARALLEL) parallel = 1;
 
 		proto = hints->ai_protocol;
 		if (hints->ai_socktype == SOCK_DGRAM)
@@ -509,9 +514,10 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 	if (family != PF_UNSPEC) na++;
 	if (canonname != 0) na++;
 	if (passive != 0) na++;
+	if (parallel != 0) na++;
 	if (numerichost != 0) na++;
 
-	if (!xdr_int(&outxdr, &na))
+	if (!xdr_int(&outxdr, (int32_t *)&na))
 	{
 		xdr_destroy(&outxdr);
 		return EAI_SYSTEM;
@@ -544,7 +550,7 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 			return EAI_SYSTEM;
 		}
 	}
-	
+
 	if (socktype != SOCK_UNSPEC)
 	{
 		snprintf(str, 64, "%u", socktype);
@@ -554,7 +560,7 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 			return EAI_SYSTEM;
 		}
 	}
-	
+
 	if (family != PF_UNSPEC)
 	{
 		snprintf(str, 64, "%u", family);
@@ -564,7 +570,7 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 			return EAI_SYSTEM;
 		}
 	}
-	
+
 	if (canonname != 0)
 	{
 		if (encode_kv(&outxdr, "canonname", "1") != 0)
@@ -573,7 +579,7 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 			return EAI_SYSTEM;
 		}
 	}
-	
+
 	if (passive != 0)
 	{
 		if (encode_kv(&outxdr, "passive", "1") != 0)
@@ -582,7 +588,16 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 			return EAI_SYSTEM;
 		}
 	}
-	
+
+	if (parallel != 0)
+	{
+		if (encode_kv(&outxdr, "parallel", "1") != 0)
+		{
+			xdr_destroy(&outxdr);
+			return EAI_SYSTEM;
+		}
+	}
+
 	if (numerichost != 0)
 	{
 		if (encode_kv(&outxdr, "numerichost", "1") != 0)
@@ -595,7 +610,7 @@ gai_make_query(const char *nodename, const char *servname, const struct addrinfo
 	*len = xdr_getpos(&outxdr);
 
 	xdr_destroy(&outxdr);
-	
+
 	return 0;
 }
 
@@ -615,10 +630,128 @@ is_a_number(const char *s)
 	return 1;
 }
 
+static int
+gai_trivial(struct in_addr *in4, struct in6_addr *in6, int16_t port, const struct addrinfo *hints, struct addrinfo **res)
+{
+	int32_t family, wantv4, wantv6, proto;
+	char *loopv4, *loopv6;
+	struct in_addr a4;
+	struct in6_addr a6;
+	struct addrinfo *a;
+
+	family = PF_UNSPEC;
+	if (hints != NULL) family = hints->ai_family;
+
+	wantv4 = 1;
+	wantv6 = 1;
+
+	if (family == PF_INET6) wantv4 = 0;
+	if (family == PF_INET) wantv6 = 0;
+
+	memset(&a4, 0, sizeof(struct in_addr));
+	memset(&a6, 0, sizeof(struct in6_addr));
+
+	if ((in4 == NULL) && (in6 == NULL))
+	{
+		loopv4 = "127.0.0.1";
+		loopv6 = "0:0:0:0:0:0:0:1";
+
+		if ((hints != NULL) && ((hints->ai_flags & AI_PASSIVE) == 1))
+		{
+			loopv4 = "0.0.0.0";
+			loopv6 = "0:0:0:0:0:0:0:0";
+		}
+
+		if ((family == PF_UNSPEC) || (family == PF_INET))
+		{
+			inet_pton(AF_INET, loopv4, &a4);
+		}
+
+		if ((family == PF_UNSPEC) || (family == PF_INET6))
+		{
+			inet_pton(AF_INET6, loopv6, &a6);
+		}
+	}
+	else if (in4 == NULL)
+	{
+		if (family == PF_INET) return EAI_BADHINTS;
+
+		wantv4 = 0;
+		memcpy(&a6, in6, sizeof(struct in6_addr));
+	}
+	else if (in6 == NULL)
+	{
+		if (family == PF_INET6) return EAI_BADHINTS;
+
+		wantv6 = 0;
+		memcpy(&a4, in4, sizeof(struct in_addr));
+	}
+	else
+	{
+		return EAI_NODATA;
+	}
+
+	proto = IPPROTO_UNSPEC;
+
+	if (hints != NULL)
+	{
+		proto = hints->ai_protocol;
+		if (proto == IPPROTO_UNSPEC)
+		{
+			if (hints->ai_socktype == SOCK_DGRAM) proto = IPPROTO_UDP;
+			else if (hints->ai_socktype == SOCK_STREAM) proto = IPPROTO_TCP;
+		}
+	}
+
+	if (wantv4 == 1)
+	{
+		if ((proto == IPPROTO_UNSPEC) || (proto == IPPROTO_UDP))
+		{
+			a = new_addrinfo_v4(0, SOCK_DGRAM, IPPROTO_UDP, port, a4, 0, NULL);
+			append_addrinfo(res, a);
+		}
+
+		if ((proto == IPPROTO_UNSPEC) || (proto == IPPROTO_TCP))
+		{
+			a = new_addrinfo_v4(0, SOCK_STREAM, IPPROTO_TCP, port, a4, 0, NULL);
+			append_addrinfo(res, a);
+		}
+
+		if (proto == IPPROTO_ICMP)
+		{
+			a = new_addrinfo_v4(0, SOCK_RAW, IPPROTO_ICMP, port, a4, 0, NULL);
+			append_addrinfo(res, a);
+		}
+	}
+
+	if (wantv6 == 1)
+	{
+		if ((proto == IPPROTO_UNSPEC) || (proto == IPPROTO_UDP))
+		{
+			a = new_addrinfo_v6(0, SOCK_DGRAM, IPPROTO_UDP, port, a6, 0, NULL);
+			append_addrinfo(res, a);
+		}
+
+		if ((proto == IPPROTO_UNSPEC) || (proto == IPPROTO_TCP))
+		{
+			a = new_addrinfo_v6(0, SOCK_STREAM, IPPROTO_TCP, port, a6, 0, NULL);
+			append_addrinfo(res, a);
+		}
+
+		if (proto == IPPROTO_ICMPV6)
+		{
+			a = new_addrinfo_v6(0, SOCK_RAW, IPPROTO_ICMPV6, port, a6, 0, NULL);
+			append_addrinfo(res, a);
+		}
+	}
+
+	return 0;
+}
+
 int
 gai_files(const char *nodename, const char *servname, const struct addrinfo *hints, struct addrinfo **res)
 {
-	int32_t i, numericserv, numerichost, family, proto, wantv4, wantv6;
+	int32_t i, status, numericserv, numerichost, family, proto, wantv4, wantv6;
 	int16_t port;
 	struct servent *s;
 	struct hostent *h;
@@ -662,14 +795,22 @@ gai_files(const char *nodename, const char *servname, const struct addrinfo *hin
 	{
 		if ((family == PF_UNSPEC) || (family == PF_INET))
 		{
-			numerichost = inet_pton(AF_INET, nodename, &a4);
-			if ((numerichost == 1) && (family == PF_UNSPEC)) family = PF_INET;
+			status = inet_pton(AF_INET, nodename, &a4);
+			if (status == 1)
+			{
+				numerichost = 1;
+				if (family == PF_UNSPEC) family = PF_INET;
+			}
 		}
 
 		if ((family == PF_UNSPEC) || (family == PF_INET6))
 		{
-			numerichost = inet_pton(AF_INET6, nodename, &a6);
-			if ((numerichost == 1) && (family == PF_UNSPEC)) family = PF_INET6;
+			status = inet_pton(AF_INET6, nodename, &a6);
+			if (status == 1)
+			{
+				numerichost = 1;
+				if (family == PF_UNSPEC) family = PF_INET6;
+			}
 		}
 	}
 
@@ -725,6 +866,12 @@ gai_files(const char *nodename, const char *servname, const struct addrinfo *hin
 				a = new_addrinfo_v4(0, SOCK_STREAM, IPPROTO_TCP, port, a4, 0, NULL);
 				append_addrinfo(res, a);
 			}
+
+			if (proto == IPPROTO_ICMP)
+			{
+				a = new_addrinfo_v4(0, SOCK_RAW, IPPROTO_ICMP, port, a4, 0, NULL);
+				append_addrinfo(res, a);
+			}
 		}
 
 		if (wantv6 == 1)
@@ -740,6 +887,12 @@ gai_files(const char *nodename, const char *servname, const struct addrinfo *hin
 				a = new_addrinfo_v6(0, SOCK_STREAM, IPPROTO_TCP, port, a6, 0, NULL);
 				append_addrinfo(res, a);
 			}
+
+			if (proto == IPPROTO_ICMPV6)
+			{
+				a = new_addrinfo_v6(0, SOCK_RAW, IPPROTO_ICMPV6, port, a6, 0, NULL);
+				append_addrinfo(res, a);
+			}
 		}
 
 		return 0;
@@ -812,7 +965,7 @@ gai_lookupd(const char *nodename, const char *servname, const struct addrinfo *h
 
 	xdrmem_create(&inxdr, rbuf, rlen, XDR_DECODE);
 
-	if (!xdr_int(&inxdr, &n))
+	if (!xdr_int(&inxdr, (int32_t *)&n))
 	{
 		xdr_destroy(&inxdr);
 		return EAI_SYSTEM;
@@ -862,9 +1015,13 @@ gai_checkhints(const struct addrinfo *hints)
 }
 
 int
-getaddrinfo(const char *nodename, const char *servname, const struct addrinfo *hints, struct addrinfo **res)
+getaddrinfo(const char * __restrict nodename, const char * __restrict servname, const struct addrinfo * __restrict hints, struct addrinfo ** __restrict res)
 {
 	int32_t status, nodenull, servnull;
+	int32_t numericserv, numerichost, family;
+	int16_t port;
+	struct in_addr a4, *p4;
+	struct in6_addr a6, *p6;
 
 	if (res == NULL) return 0;
 	*res = NULL;
@@ -875,11 +1032,60 @@ getaddrinfo(const char *nodename, const char *servname, const struct addrinfo *h
 
 	servnull = 0;
 	if ((servname == NULL) || (servname[0] == '\0')) servnull = 1;
-	
+
 	if ((nodenull == 1) && (servnull == 1)) return EAI_NONAME;
+
 	status = gai_checkhints(hints);
 	if (status != 0) return status;
 
+	/*
+	 * Trap the "trivial" cases that can be answered without a query.
+	 * (nodename == numeric) && (servname == NULL)
+	 * (nodename == numeric) && (servname == numeric)
+	 * (nodename == NULL) && (servname == numeric)
+	 */
+	p4 = NULL;
+	p6 = NULL;
+
+	memset(&a4, 0, sizeof(struct in_addr));
+	memset(&a6, 0, sizeof(struct in6_addr));
+
+	numericserv = 0;
+	port = 0;
+	if (servnull == 0) numericserv = is_a_number(servname);
+	if (numericserv == 1) port = atoi(servname);
+
+	family = PF_UNSPEC;
+	if (hints != NULL) family = hints->ai_family;
+
+	numerichost = 0;
+	if (nodenull == 0)
+	{
+		if ((family == PF_UNSPEC) || (family == PF_INET))
+		{
+			status = inet_pton(AF_INET, nodename, &a4);
+			if (status == 1)
+			{
+				p4 = &a4;
+				numerichost = 1;
+			}
+		}
+
+		if ((family == PF_UNSPEC) || (family == PF_INET6))
+		{
+			status = inet_pton(AF_INET6, nodename, &a6);
+			if (status == 1)
+			{
+				p6 = &a6;
+				numerichost = 1;
+			}
+		}
+	}
+
+	if ((nodenull == 1) && (numericserv == 1)) return gai_trivial(NULL, NULL, port, hints, res);
+	if ((numerichost == 1) && (numericserv == 1)) return gai_trivial(p4, p6, port, hints, res);
+	if ((numerichost == 1) && (servnull == 1)) return gai_trivial(p4, p6, 0, hints, res);
+
 	if (nodenull == 1) status = gai_lookupd(NULL, servname, hints, res);
 	else if (servnull == 1) status = gai_lookupd(nodename, NULL, hints, res);
 	else status = gai_lookupd(nodename, servname, hints, res);
@@ -944,7 +1150,7 @@ gai_extract_data(char *buf, uint32_t len, struct addrinfo **res)
 
 	xdrmem_create(&xdr, buf, len, XDR_DECODE);
 
-	if (!xdr_int(&xdr, &n))
+	if (!xdr_int(&xdr, (int32_t *)&n))
 	{
 		xdr_destroy(&xdr);
 		return EAI_SYSTEM;
@@ -1016,8 +1222,8 @@ getaddrinfo_async_handle_reply(void *msg)
 	if (status != KERN_SUCCESS)
 	{
 		if (status == MIG_REPLY_MISMATCH) return 0;
-		if (callback != NULL) callback(EAI_SYSTEM, NULL, context);
-		return EAI_SYSTEM;
+		if (callback != NULL) callback(EAI_NODATA, NULL, context);
+		return EAI_NODATA;
 	}
 
 	status = gai_extract_data(buf, len, &res);
@@ -1065,50 +1271,45 @@ getaddrinfo_async_handle_reply(void *msg)
 static int
 gni_lookupd_process_dictionary(XDR *inxdr, char **host, char **serv)
 {
-	int32_t i, nkeys, nvals;
-	char *key, *val;
+	int32_t i, j, nkeys, nvals, status;
+	char *key, **vals;
 
 	if ((host == NULL) || (serv == NULL)) return EAI_SYSTEM;
+
 	if (!xdr_int(inxdr, &nkeys)) return EAI_SYSTEM;
 
+	*host = NULL;
+	*serv = NULL;
+
 	for (i = 0; i < nkeys; i++)
 	{
 		key = NULL;
-		val = NULL;
+		vals = NULL;
+		nvals = 0;
 
-		if (!xdr_string(inxdr, &key, LONG_STRING_LENGTH)) return EAI_SYSTEM;
-		if (!xdr_int(inxdr, &nvals)) 
+		status = _lu_xdr_attribute(inxdr, &key, &vals, (uint32_t *)&nvals);
+		if (status < 0) return EAI_SYSTEM;
+
+		if (nvals == 0)
 		{
 			free(key);
-			return EAI_SYSTEM;
+			continue;
 		}
 
-		if (nvals != 1)
+		if ((*host == NULL) && (!strcmp("name", key)))
 		{
-			free(key);
-			return EAI_SYSTEM;
+			*host = vals[0];
+			for (j = 1; j < nvals; j++) free(vals[j]);
 		}
 
-		if (!xdr_string(inxdr, &val, LONG_STRING_LENGTH))
+		else if ((*serv == NULL) && (!strcmp(key, "service")))
 		{
-			free(key);
-			return EAI_SYSTEM;
-		}
-
-		if (!strcmp(key, "name"))
-		{
-			*host = val;
-			val = NULL;
-		}
-
-		else if (!strcmp(key, "service"))
-		{
-			*serv = val;
-			val = NULL;
+			*serv = vals[0];
+			for (j = 1; j < nvals; j++) free(vals[j]);
 		}
 
 		if (key != NULL) free(key);
-		if (val != NULL) free(val);
+		free(vals);
 	}
 
 	return 0;
@@ -1120,7 +1321,7 @@ gni_make_query(const struct sockaddr *sa, size_t salen, int wanthost, int wantse
 	XDR outxdr;
 	uint16_t port;
 	char str[_LU_MAXLUSTRLEN], *key, ifname[IF_NAMESIZE];
-	uint32_t a4, ifnum, offset, na, proto, fqdn, numerichost, numericserv, name_req;
+	uint32_t a4, ifnum, offset, na, proto, fqdn, numerichost, numericserv, name_req, isll, issl;
 	struct sockaddr_in6 *s6;
 
 	if (sa == NULL) return EAI_FAIL;
@@ -1131,6 +1332,8 @@ gni_make_query(const struct sockaddr *sa, size_t salen, int wanthost, int wantse
 	numerichost = 0;
 	numericserv = 0;
 	name_req = 0;
+	isll = 0;
+	issl = 0;
 
 	offset = INET_NTOP_AF_INET_OFFSET;
 	key = "ip_address";
@@ -1165,7 +1368,11 @@ gni_make_query(const struct sockaddr *sa, size_t salen, int wanthost, int wantse
 			default:
 				if (IN6_IS_ADDR_LINKLOCAL(&s6->sin6_addr))
 				{
-					flags |= NI_NUMERICHOST;
+					isll = 1;
+				}
+				else if (IN6_IS_ADDR_SITELOCAL(&s6->sin6_addr))
+				{
+					issl = 1;
 				}
 				else if (IN6_IS_ADDR_MULTICAST(&s6->sin6_addr))
 				{
@@ -1174,6 +1381,17 @@ gni_make_query(const struct sockaddr *sa, size_t salen, int wanthost, int wantse
 				break;
 		}
 
+		if ((isll != 0) || (issl != 0))
+		{
+			ifnum = s6->sin6_addr.__u6_addr.__u6_addr16[1];
+			if (ifnum == 0) ifnum = s6->sin6_scope_id;
+			else if ((s6->sin6_scope_id != 0) && (ifnum != s6->sin6_scope_id)) return EAI_FAIL;
+
+			s6->sin6_addr.__u6_addr.__u6_addr16[1] = 0;
+			s6->sin6_scope_id = ifnum;
+			if ((ifnum != 0) && (flags & NI_NUMERICHOST)) flags |= NI_WITHSCOPEID;
+		}
+
 		offset = INET_NTOP_AF_INET6_OFFSET;
 		key = "ipv6_address";
 		port = s6->sin6_port;
@@ -1220,7 +1438,7 @@ gni_make_query(const struct sockaddr *sa, size_t salen, int wanthost, int wantse
 
 	xdrmem_create(&outxdr, buf, *len, XDR_ENCODE);
 
-	if (!xdr_int(&outxdr, &na))
+	if (!xdr_int(&outxdr, (int32_t *)&na))
 	{
 		xdr_destroy(&outxdr);
 		return EAI_SYSTEM;
@@ -1305,17 +1523,18 @@ gni_make_query(const struct sockaddr *sa, size_t salen, int wanthost, int wantse
 	*len = xdr_getpos(&outxdr);
 
 	xdr_destroy(&outxdr);
-	
+
 	return 0;
 }
 
 int
-getnameinfo(const struct sockaddr *sa, socklen_t salen, char *host, size_t hostlen, char *serv, size_t servlen, int flags)
+getnameinfo(const struct sockaddr * __restrict sa, socklen_t salen, char * __restrict host, socklen_t hostlen, char * __restrict serv, socklen_t servlen, int flags)
 {
 	uint32_t n, i, qlen, rlen;
-	int wanth, wants;
+	uint32_t ifnum;
+	int wanth, wants, isll, issl;
 	XDR inxdr;
-	char qbuf[LU_QBUF_SIZE];
+	char qbuf[LU_QBUF_SIZE], ifname[IF_NAMESIZE];
 	char *rbuf, *hval, *sval;
 	mach_port_t server_port;
 	kern_return_t status;
@@ -1325,10 +1544,34 @@ getnameinfo(const struct sockaddr *sa, socklen_t salen, char *host, size_t hostl
 	/* Check input */
 	if (sa == NULL) return EAI_FAIL;
 
-	/* V4 mapped and compat addresses are converted to plain V4 */
+	isll = 0;
+	issl = 0;
+	ifnum = 0;
+
 	if (sa->sa_family == AF_INET6)
 	{
 		s6 = (struct sockaddr_in6 *)sa;
+
+		if (IN6_IS_ADDR_LINKLOCAL(&s6->sin6_addr)) isll = 1;
+		if (IN6_IS_ADDR_SITELOCAL(&s6->sin6_addr)) issl = 1;
+
+		/*
+		 * Link-local and site-local IPv6 addresses may have a scope id 
+		 * in s6->sin6_addr.__u6_addr.__u6_addr16[1] as well as in s6->sin6_scope_id.
+		 * If they are both non-zero, they must be equal.
+		 * We zero s6->sin6_addr.__u6_addr.__u6_addr16[1] and set s6->sin6_scope_id.
+		 */
+		if ((isll != 0) || (issl != 0))
+		{
+			ifnum = ntohs(s6->sin6_addr.__u6_addr.__u6_addr16[1]);
+			if (ifnum == 0) ifnum = s6->sin6_scope_id;
+			else if ((s6->sin6_scope_id != 0) && (ifnum != s6->sin6_scope_id)) return EAI_FAIL;
+
+			s6->sin6_addr.__u6_addr.__u6_addr16[1] = 0;
+			s6->sin6_scope_id = ifnum;
+		}
+
+		/* V4 mapped and compat addresses are converted to plain V4 */
 		if ((IN6_IS_ADDR_V4MAPPED(&s6->sin6_addr)) || (IN6_IS_ADDR_V4COMPAT(&s6->sin6_addr)))
 		{
 			s4 = (struct sockaddr_in *)calloc(1, sizeof(struct sockaddr_in));
@@ -1360,6 +1603,16 @@ getnameinfo(const struct sockaddr *sa, socklen_t salen, char *host, size_t hostl
 		if (sa->sa_family == PF_INET6) i = INET_NTOP_AF_INET6_OFFSET;
 		if (inet_ntop(sa->sa_family, (char *)(sa) + i, host, hostlen) == NULL) return EAI_FAIL;
 
+		if (((isll != 0) || (issl != 0)) && (ifnum != 0))
+		{
+			/* append interface name */
+			if (if_indextoname(ifnum, ifname) != NULL)
+			{
+				strcat(host, "%");
+				strcat(host, ifname);
+			}
+		}
+
 		if (wants == 0) return 0;
 	}
 
@@ -1399,7 +1652,7 @@ getnameinfo(const struct sockaddr *sa, socklen_t salen, char *host, size_t hostl
 
 	qlen = LU_QBUF_SIZE;
 	i = gni_make_query(sa, salen, wanth, wants, flags, qbuf, &qlen);
-	if (i != 0) return EAI_SYSTEM;
+	if (i != 0) return i;
 
 	qlen /= BYTES_PER_XDR_UNIT;
 
@@ -1412,7 +1665,7 @@ getnameinfo(const struct sockaddr *sa, socklen_t salen, char *host, size_t hostl
 
 	xdrmem_create(&inxdr, rbuf, rlen, XDR_DECODE);
 
-	if (!xdr_int(&inxdr, &n))
+	if (!xdr_int(&inxdr, (int32_t *)&n))
 	{
 		xdr_destroy(&inxdr);
 		return EAI_SYSTEM;
@@ -1478,7 +1731,7 @@ getnameinfo_async_start(mach_port_t *p, const struct sockaddr *sa, size_t salen,
 
 	qlen = LU_QBUF_SIZE;
 	i = gni_make_query(sa, salen, 1, 1, flags, qbuf, &qlen);
-	if (i != 0) return EAI_SYSTEM;
+	if (i != 0) return i;
 
 	qlen /= BYTES_PER_XDR_UNIT;
 
@@ -1496,6 +1749,7 @@ gni_extract_data(char *buf, uint32_t len, char **host, char **serv)
 {
 	XDR xdr;
 	uint32_t n;
+	int i;
 
 	*host = NULL;
 	*serv = NULL;
@@ -1505,7 +1759,7 @@ gni_extract_data(char *buf, uint32_t len, char **host, char **serv)
 
 	xdrmem_create(&xdr, buf, len, XDR_DECODE);
 
-	if (!xdr_int(&xdr, &n))
+	if (!xdr_int(&xdr, (int32_t *)&n))
 	{
 		xdr_destroy(&xdr);
 		return EAI_SYSTEM;
@@ -1517,7 +1771,9 @@ gni_extract_data(char *buf, uint32_t len, char **host, char **serv)
 		return EAI_NONAME;
 	}
 
-	return gni_lookupd_process_dictionary(&xdr, host, serv);
+	i = gni_lookupd_process_dictionary(&xdr, host, serv);
+	xdr_destroy(&xdr);
+	return i;
 }
 
 int32_t
@@ -1558,8 +1814,8 @@ getnameinfo_async_handle_reply(void *msg)
 	if (status != KERN_SUCCESS)
 	{
 		if (status == MIG_REPLY_MISMATCH) return 0;
-		if (callback != NULL) callback(EAI_SYSTEM, NULL, NULL, context);
-		return EAI_SYSTEM;
+		if (callback != NULL) callback(EAI_NONAME, NULL, NULL, context);
+		return EAI_NONAME;
 	}
 
 	hval = NULL;
