@@ -41,7 +41,7 @@
 #define LU_MESSAGE_SEND_ID 4241776
 #define LU_MESSAGE_REPLY_ID 4241876
 
-static pthread_key_t _info_key = NULL;
+static pthread_key_t _info_key = 0;
 static pthread_once_t _info_key_initialized = PTHREAD_ONCE_INIT;
 
 struct _lu_data_s
@@ -112,9 +112,9 @@ _lu_async_send(_lu_async_request_t *r)
 	inp->query_data_len = r->request_buffer_len;
 	memcpy(inp->query_data, r->request_buffer, 4 * r->request_buffer_len);
 
-	status = mach_msg(&inp->head, MACH_SEND_MSG, msgh_size, 0, MACH_PORT_NULL, MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);	
+	status = mach_msg(&inp->head, MACH_SEND_MSG, msgh_size, 0, MACH_PORT_NULL, MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);
 	if (status == MACH_MSG_SUCCESS) return KERN_SUCCESS;
-	
+
 	if (status == MACH_SEND_INVALID_REPLY)
 	{
 		mach_port_mod_refs(mach_task_self(), r->reply_port, MACH_PORT_RIGHT_RECEIVE, -1);
@@ -204,6 +204,7 @@ lu_async_receive(mach_port_t p, char **buf, uint32_t *len)
 	kern_return_t status;
 	uint32_t size;
 	_lu_async_request_t *req;
+	boolean_t msgh_simple;
 
 	size = sizeof(_lu_reply_msg_t);
 
@@ -219,6 +220,8 @@ lu_async_receive(mach_port_t p, char **buf, uint32_t *len)
 		return status;
 	}
 
+	msgh_simple = !(r->head.msgh_bits & MACH_MSGH_BITS_COMPLEX);
+
 	req = _lu_worklist_remove(r->head.msgh_local_port);
 	if (req == NULL)
 	{
@@ -226,6 +229,14 @@ lu_async_receive(mach_port_t p, char **buf, uint32_t *len)
 		return KERN_FAILURE;
 	}
 
+	if (msgh_simple && ((mig_reply_error_t *) r)->RetCode != KERN_SUCCESS)
+	{
+		_lu_free_request(req);
+		status = ((mig_reply_error_t *) r)->RetCode;
+		free(r);
+		return status;
+	}
+
 	*buf = r->reply_data.address;
 	*len = r->reply_data.size;
 
@@ -273,7 +284,7 @@ _lu_create_request(uint32_t proc, const char *buf, uint32_t len, void *callback,
 	_lu_async_request_t *r;
 	kern_return_t status;
 
-	if (_lu_port == NULL) return NULL;
+	if (_lu_port == MACH_PORT_NULL) return NULL;
 
 	r = (_lu_async_request_t *)calloc(1, sizeof(_lu_async_request_t));
 	if (r == NULL) return NULL;
@@ -292,6 +303,12 @@ _lu_create_request(uint32_t proc, const char *buf, uint32_t len, void *callback,
 	r->proc = proc;
 
 	r->request_buffer = malloc(len * BYTES_PER_XDR_UNIT);
+	if (r->request_buffer == NULL)
+	{
+		free(r);
+		return NULL;
+	}
+
 	memcpy(r->request_buffer, buf, len * BYTES_PER_XDR_UNIT);
 	r->request_buffer_len = len;
 
@@ -350,6 +367,7 @@ lu_async_handle_reply(void *msg, char **buf, uint32_t *len, void **callback, voi
 	_lu_async_request_t *req;
 	kern_return_t status;
 	uint32_t retry;
+	boolean_t msgh_simple;
 
 	if (msg == NULL) return -1;
 	r = (_lu_reply_msg_t *)msg;
@@ -375,15 +393,23 @@ lu_async_handle_reply(void *msg, char **buf, uint32_t *len, void **callback, voi
 		return MIG_REPLY_MISMATCH;
 	}
 
+        msgh_simple = !(r->head.msgh_bits & MACH_MSGH_BITS_COMPLEX);
+
 	req = _lu_worklist_remove(r->head.msgh_local_port);
 	if (req == NULL) return -1;
 
-	*buf = r->reply_data.address;
-	*len = r->reply_data.size;
 	*callback = req->callback;
 	*context = req->context;
-
 	_lu_free_request(req);
+ 
+	if (msgh_simple && ((mig_reply_error_t *) r)->RetCode != KERN_SUCCESS)
+	{
+		return ((mig_reply_error_t *) r)->RetCode;
+	}
+
+	*buf = r->reply_data.address;
+	*len = r->reply_data.size;
+
 	return 0;
 }
 
@@ -397,14 +423,20 @@ _lookupd_xdr_dictionary(XDR *inxdr)
 	if (!xdr_int(inxdr, &nkeys)) return NULL;
 
 	l = (ni_proplist *)malloc(sizeof(ni_proplist));
+	if (l == NULL) return NULL;
+
 	NI_INIT(l);
 
 	l->ni_proplist_len = nkeys;
 	l->ni_proplist_val = NULL;
 	if (nkeys > 0)
 	{
-		i = nkeys * sizeof(ni_property);
-		l->ni_proplist_val = (ni_property *)calloc(1, i);
+		l->ni_proplist_val = (ni_property *)calloc(nkeys, sizeof(ni_property));
+		if (l->ni_proplist_val == NULL)
+		{
+			free(l);
+			return NULL;
+		}
 	}
 
 	for (i = 0; i < nkeys; i++)
@@ -417,20 +449,24 @@ _lookupd_xdr_dictionary(XDR *inxdr)
 		}
 
 		l->ni_proplist_val[i].nip_name = key;
-	
+
 		if (!xdr_int(inxdr, &nvals))
 		{
 			ni_proplist_free(l);
 			return NULL;
 		}
-	
+
 		l->ni_proplist_val[i].nip_val.ni_namelist_len = nvals;
 		if (nvals > 0)
 		{
-			j = nvals * sizeof(ni_name);
-			l->ni_proplist_val[i].nip_val.ni_namelist_val = (ni_name *)calloc(1, j);
+			l->ni_proplist_val[i].nip_val.ni_namelist_val = (ni_name *)calloc(nvals, sizeof(ni_name));
+			if (l->ni_proplist_val[i].nip_val.ni_namelist_val == NULL)
+			{
+				ni_proplist_free(l);
+				return NULL;
+			}
 		}
-		
+
 		for (j = 0; j < nvals; j++)
 		{
 			val = NULL;
@@ -463,7 +499,7 @@ lookupd_query(ni_proplist *l, ni_proplist ***out)
 	if (l == NULL) return 0;
 	if (out == NULL) return 0;
 
-	if (_lu_port == NULL) return 0;
+	if (_lu_port == MACH_PORT_NULL) return 0;
 
 	status = _lookup_link(_lu_port, "query", &proc);
 	if (status != KERN_SUCCESS) return 0;
@@ -482,13 +518,12 @@ lookupd_query(ni_proplist *l, ni_proplist ***out)
 	for (i = 0; i < l->ni_proplist_len; i++)
 	{
 		p = &(l->ni_proplist_val[i]);
-		s = NULL;
+		s = p->nip_name;
 		if (!xdr_string(&outxdr, &s, _LU_MAXLUSTRLEN))
 		{
 			xdr_destroy(&outxdr);
 			return 0;
 		}
-		p->nip_name = s;
 
 		if (!xdr_int(&outxdr, &(p->nip_val.ni_namelist_len)))
 		{
@@ -498,21 +533,20 @@ lookupd_query(ni_proplist *l, ni_proplist ***out)
 
 		for (j = 0; j < p->nip_val.ni_namelist_len; j++)
 		{
-			s = NULL;
+			s = p->nip_val.ni_namelist_val[j];
 			if (!xdr_string(&outxdr, &s, _LU_MAXLUSTRLEN))
 			{
 				xdr_destroy(&outxdr);
 				return 0;
 			}
-			p->nip_val.ni_namelist_val[j] = s;
 		}
 	}
 
 	listbuf = NULL;
 	datalen = 0;
 
-	n = xdr_getpos(&outxdr) / BYTES_PER_XDR_UNIT;
-	status = _lookup_all(_lu_port, proc, (unit *)databuf, n, &listbuf, &datalen);
+	n = xdr_getpos(&outxdr);
+	status = _lookup_all(_lu_port, proc, (void *)databuf, n, &listbuf, &datalen);
 	if (status != KERN_SUCCESS)
 	{
 		xdr_destroy(&outxdr);
@@ -520,12 +554,7 @@ lookupd_query(ni_proplist *l, ni_proplist ***out)
 	}
 
 	xdr_destroy(&outxdr);
-
-#ifdef NOTDEF
-/* NOTDEF because OOL buffers are counted in bytes with untyped IPC */
 	datalen *= BYTES_PER_XDR_UNIT;
-#endif
-
 	xdrmem_create(&inxdr, listbuf, datalen, XDR_DECODE);
 
 	if (!xdr_int(&inxdr, &n))
@@ -541,6 +570,11 @@ lookupd_query(ni_proplist *l, ni_proplist ***out)
 	}
 
 	*out = (ni_proplist **)malloc(n * sizeof(ni_proplist *));
+	if (out == NULL)
+	{
+		xdr_destroy(&inxdr);
+		return 0;
+	}
 
 	for (i = 0; i < n; i++)
 	{
@@ -550,7 +584,7 @@ lookupd_query(ni_proplist *l, ni_proplist ***out)
 	xdr_destroy(&inxdr);
 
 	vm_deallocate(mach_task_self(), (vm_address_t)listbuf, datalen);
-	
+
 	return n;
 }
 
@@ -567,6 +601,8 @@ lookupd_make_query(char *cat, char *fmt, ...)
 	if (fmt[0] != 'k') return NULL;
 
 	l = (ni_proplist *)malloc(sizeof(ni_proplist));
+	if (l == NULL) return NULL;
+
 	NI_INIT(l);
 
 	na = 0;
@@ -575,27 +611,61 @@ lookupd_make_query(char *cat, char *fmt, ...)
 	if (cat != NULL)
 	{
 		l->ni_proplist_val = (ni_property *)malloc(sizeof(ni_property));
+		if (l->ni_proplist_val == NULL)
+		{
+			free(l);
+			return NULL;
+		}
+
 		p = &(l->ni_proplist_val[0]);
 		arg = "_lookup_category";
 		p->nip_name = strdup(arg);
+		if (p->nip_name == NULL)
+		{
+			ni_proplist_free(l);
+			return NULL;
+		}
+		
 		p->nip_val.ni_namelist_len = 1;
 		p->nip_val.ni_namelist_val = (ni_name *)malloc(sizeof(ni_name));
+		if (p->nip_val.ni_namelist_val == NULL)
+		{
+			ni_proplist_free(l);
+			return NULL;
+		}
+			
 		p->nip_val.ni_namelist_val[0] = strdup(cat);
-
+		if (p->nip_val.ni_namelist_val[0] == NULL)
+		{
+			ni_proplist_free(l);
+			return NULL;
+		}
+		
 		l->ni_proplist_len++;
 		x++;
 	}
 
 	va_start(ap, fmt);
-	for (f = fmt; *f != NULL; f++)
+	for (f = fmt; (*f) != '\0'; f++)
 	{
 		arg = va_arg(ap, char *);
 		if (*f == 'k')
 		{
-			l->ni_proplist_val = (ni_property *)realloc(l->ni_proplist_val, (l->ni_proplist_len + 1) * sizeof(ni_property));
-
+			l->ni_proplist_val = (ni_property *)reallocf(l->ni_proplist_val, (l->ni_proplist_len + 1) * sizeof(ni_property));
+			if (l->ni_proplist_val == NULL)
+			{
+				ni_proplist_free(l);
+				return NULL;
+			}
+			
 			p = &(l->ni_proplist_val[l->ni_proplist_len]);
 			p->nip_name = strdup(arg);
+			if (p->nip_name == NULL)
+			{
+				ni_proplist_free(l);
+				return NULL;
+			}
+
 			p->nip_val.ni_namelist_len = 0;
 			p->nip_val.ni_namelist_val = NULL;
 
@@ -611,9 +681,22 @@ lookupd_make_query(char *cat, char *fmt, ...)
 			}
 			else
 			{
-				p->nip_val.ni_namelist_val = (ni_name *)realloc(p->nip_val.ni_namelist_val, (p->nip_val.ni_namelist_len + 1) * sizeof(ni_name));
+				p->nip_val.ni_namelist_val = (ni_name *)reallocf(p->nip_val.ni_namelist_val, (p->nip_val.ni_namelist_len + 1) * sizeof(ni_name));
 			}
+
+			if (p->nip_val.ni_namelist_val == NULL)
+			{
+				ni_proplist_free(l);
+				return NULL;
+			}
+
 			p->nip_val.ni_namelist_val[p->nip_val.ni_namelist_len] = strdup(arg);
+			if (p->nip_val.ni_namelist_val[p->nip_val.ni_namelist_len] == NULL)
+			{
+				ni_proplist_free(l);
+				return NULL;
+			}
+
 			p->nip_val.ni_namelist_len++;
 		}
 	}
@@ -640,8 +723,17 @@ ni_property_merge(ni_property *a, ni_property *b)
 
 		if (addme == 1)
 		{
-			a->nip_val.ni_namelist_val = (ni_name *)realloc(a->nip_val.ni_namelist_val, (a->nip_val.ni_namelist_len + 1) * sizeof(ni_name));
+			a->nip_val.ni_namelist_val = (ni_name *)reallocf(a->nip_val.ni_namelist_val, (a->nip_val.ni_namelist_len + 1) * sizeof(ni_name));
+			if (a->nip_val.ni_namelist_val == NULL) return;
+
 			a->nip_val.ni_namelist_val[a->nip_val.ni_namelist_len] = strdup(b->nip_val.ni_namelist_val[j]);
+			if (a->nip_val.ni_namelist_val[a->nip_val.ni_namelist_len] == NULL)
+			{
+				free(a->nip_val.ni_namelist_val);
+				a->nip_val.ni_namelist_val = NULL;
+				return;
+			}
+
 			a->nip_val.ni_namelist_len++;
 		}
 	}
@@ -665,8 +757,17 @@ ni_proplist_merge(ni_proplist *a, ni_proplist *b)
 		}
 		if (addme == 1)
 		{
-			a->ni_proplist_val = (ni_property *)realloc(a->ni_proplist_val, (a->ni_proplist_len + 1) * sizeof(ni_property));
+			a->ni_proplist_val = (ni_property *)reallocf(a->ni_proplist_val, (a->ni_proplist_len + 1) * sizeof(ni_property));
+			if (a->ni_proplist_val == NULL) return;
+
 			a->ni_proplist_val[a->ni_proplist_len].nip_name = strdup(b->ni_proplist_val[wb].nip_name);
+			if (a->ni_proplist_val[a->ni_proplist_len].nip_name == NULL)
+			{
+				free(a->ni_proplist_val);
+				a->ni_proplist_val = NULL;
+				return NULL;
+			}
+	
 			a->ni_proplist_val[a->ni_proplist_len].nip_val.ni_namelist_len = 0;
 			a->ni_proplist_val[a->ni_proplist_len].nip_val.ni_namelist_val = NULL;
 			a->ni_proplist_len++;
@@ -696,7 +797,7 @@ _lu_data_free(void *x)
 	t = (struct _lu_data_s *)x;
 
 	for (i = 0; i < t->icount; i++)
-	{		
+	{
 		if ((t->idata[i] != NULL) && (t->idata_destructor[i] != NULL))
 		{
 			(*(t->idata_destructor[i]))(t->idata[i]);
@@ -740,6 +841,7 @@ _lu_data_get()
 	if (libinfo_data != NULL) return libinfo_data;
 
 	libinfo_data = (struct _lu_data_s *)calloc(1, sizeof(struct _lu_data_s));
+	if (libinfo_data == NULL) return NULL;
 
 	pthread_setspecific(_info_key, libinfo_data);
 	return libinfo_data;
@@ -752,6 +854,7 @@ _lu_data_create_key(unsigned int key, void (*destructor)(void *))
 	unsigned int i, n;
 
 	libinfo_data = _lu_data_get();
+	if (libinfo_data == NULL) return NULL;
 
 	for (i = 0; i < libinfo_data->icount; i++)
 	{
@@ -769,9 +872,17 @@ _lu_data_create_key(unsigned int key, void (*destructor)(void *))
 	}
 	else
 	{
-		libinfo_data->ikey = (unsigned int *)realloc(libinfo_data->ikey, n * sizeof(unsigned int));
-		libinfo_data->idata = (void **)realloc(libinfo_data->idata, n * sizeof(void *));
-		libinfo_data->idata_destructor = (void (**)(void *))realloc(libinfo_data->idata_destructor, n * sizeof(void (*)(void *)));
+		libinfo_data->ikey = (unsigned int *)reallocf(libinfo_data->ikey, n * sizeof(unsigned int));
+		libinfo_data->idata = (void **)reallocf(libinfo_data->idata, n * sizeof(void *));
+		libinfo_data->idata_destructor = (void (**)(void *))reallocf(libinfo_data->idata_destructor, n * sizeof(void (*)(void *)));
+	}
+
+	if ((libinfo_data->ikey == NULL) || (libinfo_data->idata == NULL) || (libinfo_data->idata_destructor == NULL))
+	{
+		if (libinfo_data->ikey != NULL) free(libinfo_data->ikey);
+		if (libinfo_data->idata != NULL) free(libinfo_data->idata);
+		if (libinfo_data->idata_destructor != NULL) free(libinfo_data->idata_destructor);
+		return NULL;
 	}
 
 	libinfo_data->ikey[i] = key;
@@ -804,6 +915,7 @@ _lu_data_set_key(unsigned int key, void *data)
 	unsigned int i;
 
 	libinfo_data = _lu_data_get();
+	if (libinfo_data == NULL) return;
 
 	i = _lu_data_index(key, libinfo_data);
 	if (i == (unsigned int)-1) return;
@@ -818,6 +930,7 @@ _lu_data_get_key(unsigned int key)
 	unsigned int i;
 
 	libinfo_data = _lu_data_get();
+	if (libinfo_data == NULL) return NULL;
 
 	i = _lu_data_index(key, libinfo_data);
 	if (i == (unsigned int)-1) return NULL;
@@ -874,6 +987,8 @@ _lu_xdr_attribute(XDR *xdr, char **key, char ***val, unsigned int *count)
 	*count = len;
 
 	x = (char **)calloc(len + 1, sizeof(char *));
+	if (x == NULL) return -1;
+
 	*val = x;
 
 	for (i = 0; i < len; i++)
